from __future__ import annotations

import itertools
from pathlib import Path
from textwrap import dedent
from typing import TYPE_CHECKING

import pytest
from jinja2 import FileSystemLoader

from commitizen import git
from commitizen.commands.changelog import Changelog
from commitizen.exceptions import (
    DryRunExit,
    InvalidCommandArgumentError,
    NoCommitsFoundError,
    NoRevisionError,
    NotAGitProjectError,
    NotAllowed,
)

if TYPE_CHECKING:
    from pytest_mock import MockFixture
    from pytest_regressions.file_regression import FileRegressionFixture

    from commitizen.changelog_formats import ChangelogFormat
    from commitizen.config.base_config import BaseConfig
    from commitizen.config.json_config import JsonConfig
    from commitizen.cz.base import BaseCommitizen
    from tests.utils import UtilFixture


@pytest.fixture
def changelog_jinja_file(tmp_commitizen_project: Path) -> Path:
    return tmp_commitizen_project / "changelog.jinja"


@pytest.fixture
def changelog_tpl(
    tmp_commitizen_project: Path, any_changelog_format: ChangelogFormat
) -> Path:
    return tmp_commitizen_project / any_changelog_format.template


@pytest.fixture
def changelog_file(
    tmp_commitizen_project: Path, any_changelog_format: ChangelogFormat
) -> Path:
    return tmp_commitizen_project / any_changelog_format.default_changelog_file


@pytest.mark.usefixtures("tmp_commitizen_project")
def test_changelog_from_version_zero_point_two(
    capsys: pytest.CaptureFixture,
    file_regression: FileRegressionFixture,
    util: UtilFixture,
):
    util.create_file_and_commit("feat: new file")
    util.create_file_and_commit("refactor: not in changelog")

    # create tag
    util.run_cli("bump", "--yes")
    capsys.readouterr()

    util.create_file_and_commit("feat: after 0.2.0")
    util.create_file_and_commit("feat: after 0.2")

    with pytest.raises(DryRunExit):
        util.run_cli("changelog", "--dry-run", "--start-rev", "0.2.0")

    out, _ = capsys.readouterr()
    file_regression.check(out, extension=".md")


@pytest.mark.usefixtures("tmp_commitizen_project")
def test_changelog_with_different_cz(
    capsys: pytest.CaptureFixture,
    file_regression: FileRegressionFixture,
    util: UtilFixture,
):
    util.create_file_and_commit("JRA-34 #comment corrected indent issue")
    util.create_file_and_commit("JRA-35 #time 1w 2d 4h 30m Total work logged")

    with pytest.raises(DryRunExit):
        util.run_cli("-n", "cz_jira", "changelog", "--dry-run")
    out, _ = capsys.readouterr()
    file_regression.check(out, extension=".md")


@pytest.mark.usefixtures("tmp_commitizen_project")
def test_changelog_from_start(
    changelog_format: ChangelogFormat,
    file_regression: FileRegressionFixture,
    util: UtilFixture,
):
    util.create_file_and_commit("feat: new file")
    util.create_file_and_commit("refactor: is in changelog")
    util.create_file_and_commit("Merge into master")
    changelog_file = f"CHANGELOG.{changelog_format.extension}"
    template = f"CHANGELOG.{changelog_format.extension}.j2"

    util.run_cli("changelog", "--file-name", changelog_file, "--template", template)

    out = Path(changelog_file).read_text(encoding="utf-8")
    file_regression.check(out, extension=changelog_format.ext)


@pytest.mark.usefixtures("tmp_commitizen_project")
@pytest.mark.freeze_time("2022-08-14")
def test_changelog_replacing_unreleased_using_incremental(
    changelog_format: ChangelogFormat,
    file_regression: FileRegressionFixture,
    util: UtilFixture,
):
    util.create_file_and_commit("feat: add new output")
    util.create_file_and_commit("fix: output glitch")
    util.create_file_and_commit("Merge into master")
    changelog_file = f"CHANGELOG.{changelog_format.extension}"
    template = f"CHANGELOG.{changelog_format.extension}.j2"

    util.run_cli("changelog", "--file-name", changelog_file, "--template", template)

    util.run_cli("bump", "--yes", "--file-name", changelog_file, "--template", template)

    util.create_file_and_commit("fix: mama gotta work")
    util.create_file_and_commit("feat: add more stuff")
    util.create_file_and_commit("Merge into master")

    util.run_cli(
        "changelog",
        "--incremental",
        "--file-name",
        changelog_file,
        "--template",
        template,
    )

    out = Path(changelog_file).read_text(encoding="utf-8")

    file_regression.check(out, extension=changelog_format.ext)


@pytest.mark.usefixtures("tmp_commitizen_project")
@pytest.mark.freeze_time("2022-08-14")
def test_changelog_is_persisted_using_incremental(
    changelog_path: Path,
    file_regression: FileRegressionFixture,
    util: UtilFixture,
):
    util.create_file_and_commit("feat: add new output")
    util.create_file_and_commit("fix: output glitch")
    util.create_file_and_commit("Merge into master")

    util.run_cli("bump", "--yes")

    util.run_cli("changelog")

    with changelog_path.open("a", encoding="utf-8") as f:
        f.write("\nnote: this should be persisted using increment\n")

    util.create_file_and_commit("fix: mama gotta work")
    util.create_file_and_commit("feat: add more stuff")
    util.create_file_and_commit("Merge into master")

    util.run_cli("changelog", "--incremental")

    out = changelog_path.read_text(encoding="utf-8")

    file_regression.check(out, extension=".md")


@pytest.mark.usefixtures("tmp_commitizen_project")
def test_changelog_incremental_angular_sample(
    changelog_path: Path,
    file_regression: FileRegressionFixture,
    util: UtilFixture,
):
    with changelog_path.open("w", encoding="utf-8") as f:
        f.write(
            "# [10.0.0-rc.3](https://github.com/angular/angular/compare/10.0.0-rc.2...10.0.0-rc.3) (2020-04-22)\n"
            "\n"
            "### Bug Fixes"
            "\n"
            "* **common:** format day-periods that cross midnight ([#36611](https://github.com/angular/angular/issues/36611)) ([c6e5fc4](https://github.com/angular/angular/commit/c6e5fc4)), closes [#36566](https://github.com/angular/angular/issues/36566)\n"
        )
    util.create_file_and_commit("irrelevant commit")
    util.create_tag("10.0.0-rc.3")

    util.create_file_and_commit("feat: add new output")
    util.create_file_and_commit("fix: output glitch")
    util.create_file_and_commit("fix: mama gotta work")
    util.create_file_and_commit("feat: add more stuff")
    util.create_file_and_commit("Merge into master")

    util.run_cli("changelog", "--incremental")

    out = changelog_path.read_text(encoding="utf-8")

    file_regression.check(out, extension=".md")


KEEP_A_CHANGELOG = """# Changelog
All notable changes to this project will be documented in this file.

The format is based on [Keep a Changelog](https://keepachangelog.com/en/1.0.0/),
and this project adheres to [Semantic Versioning](https://semver.org/spec/v2.0.0.html).

## [Unreleased]

## [1.0.0] - 2017-06-20
### Added
- New visual identity by [@tylerfortune8](https://github.com/tylerfortune8).
- Version navigation.

### Changed
- Start using "changelog" over "change log" since it's the common usage.

### Removed
- Section about "changelog" vs "CHANGELOG".

## [0.3.0] - 2015-12-03
### Added
- RU translation from [@aishek](https://github.com/aishek).
"""


@pytest.mark.usefixtures("tmp_commitizen_project")
def test_changelog_incremental_keep_a_changelog_sample(
    changelog_path: Path,
    file_regression: FileRegressionFixture,
    util: UtilFixture,
):
    with changelog_path.open("w", encoding="utf-8") as f:
        f.write(KEEP_A_CHANGELOG)
    util.create_file_and_commit("irrelevant commit")
    util.create_tag("1.0.0")

    util.create_file_and_commit("feat: add new output")
    util.create_file_and_commit("fix: output glitch")
    util.create_file_and_commit("fix: mama gotta work")
    util.create_file_and_commit("feat: add more stuff")
    util.create_file_and_commit("Merge into master")

    util.run_cli("changelog", "--incremental")

    out = changelog_path.read_text(encoding="utf-8")

    file_regression.check(out, extension=".md")


@pytest.mark.usefixtures("tmp_commitizen_project")
@pytest.mark.parametrize("dry_run", [True, False])
def test_changelog_hook(
    mocker: MockFixture, config: BaseConfig, dry_run: bool, util: UtilFixture
):
    changelog_hook_mock = mocker.Mock()
    changelog_hook_mock.return_value = "cool changelog hook"

    util.create_file_and_commit("feat: new file")
    util.create_file_and_commit("refactor: is in changelog")
    util.create_file_and_commit("Merge into master")

    config.settings["change_type_order"] = ["Refactor", "Feat"]  # type: ignore[typeddict-unknown-key]
    changelog = Changelog(
        config, {"unreleased_version": None, "incremental": True, "dry_run": dry_run}
    )
    mocker.patch.object(changelog.cz, "changelog_hook", changelog_hook_mock)
    if dry_run:
        with pytest.raises(DryRunExit):
            changelog()
    else:
        changelog()

    full_changelog = (
        "## Unreleased\n\n### Refactor\n\n- is in changelog\n\n### Feat\n\n- new file\n"
    )
    partial_changelog = full_changelog
    if dry_run:
        partial_changelog = ""

    changelog_hook_mock.assert_called_with(full_changelog, partial_changelog)


@pytest.mark.usefixtures("tmp_commitizen_project")
def test_changelog_hook_customize(
    mocker: MockFixture, config_customize: JsonConfig, util: UtilFixture
):
    changelog_hook_mock = mocker.Mock()
    changelog_hook_mock.return_value = "cool changelog hook"

    util.create_file_and_commit("feat: new file")
    util.create_file_and_commit("refactor: is in changelog")
    util.create_file_and_commit("Merge into master")

    changelog = Changelog(
        config_customize,
        {"unreleased_version": None, "incremental": True, "dry_run": False},
    )
    mocker.patch.object(changelog.cz, "changelog_hook", changelog_hook_mock)
    changelog()
    full_changelog = (
        "## Unreleased\n\n### Refactor\n\n- is in changelog\n\n### Feat\n\n- new file\n"
    )

    changelog_hook_mock.assert_called_with(full_changelog, full_changelog)


@pytest.mark.usefixtures("tmp_commitizen_project")
def test_changelog_release_hook(
    mocker: MockFixture, config: BaseConfig, util: UtilFixture
):
    def changelog_release_hook(release: dict, tag: git.GitTag) -> dict:
        return release

    for i in range(3):
        util.create_file_and_commit("feat: new file")
        util.create_file_and_commit("refactor: is in changelog")
        util.create_file_and_commit("Merge into master")
        util.create_tag(f"0.{i + 1}.0")

    # changelog = Changelog(config, {})
    changelog = Changelog(
        config, {"unreleased_version": None, "incremental": True, "dry_run": False}
    )
    mocker.patch.object(changelog.cz, "changelog_release_hook", changelog_release_hook)
    spy = mocker.spy(changelog.cz, "changelog_release_hook")
    changelog()

    assert spy.call_count == 3


@pytest.mark.usefixtures("tmp_commitizen_project")
def test_changelog_with_non_linear_merges_commit_order(
    mocker: MockFixture, config_customize: JsonConfig, util: UtilFixture
):
    """Test that commits merged non-linearly are correctly ordered in the changelog

    A typical scenario is having two branches from main like so:
    * feat: I will be merged first - (2023-03-01 11:35:51 +0100) |  (branchB)
    | * feat: I will be merged second - (2023-03-01 11:35:22 +0100) |  (branchA)
    |/
    * feat: initial commit - (2023-03-01 11:34:54 +0100) |  (HEAD -> main)

    And merging them, for example in the reverse order they were created on would give the following:
    *   Merge branch 'branchA' - (2023-03-01 11:42:59 +0100) |  (HEAD -> main)
    |\
    | * feat: I will be merged second - (2023-03-01 11:35:22 +0100) |  (branchA)
    * | feat: I will be merged first - (2023-03-01 11:35:51 +0100) |  (branchB)
    |/
    * feat: initial commit - (2023-03-01 11:34:54 +0100) |

    In this case we want the changelog to reflect the topological order of commits,
    i.e. the order in which they were merged into the main branch

    So the above example should result in the following:
    ## Unreleased

    ### Feat
    - I will be merged second
    - I will be merged first
    - initial commit
    """
    changelog_hook_mock = mocker.Mock()
    changelog_hook_mock.return_value = "cool changelog hook"

    util.create_file_and_commit("feat: initial commit")

    main_branch = util.get_current_branch()

    util.create_branch("branchA")
    util.create_branch("branchB")

    util.switch_branch("branchA")
    util.create_file_and_commit("feat: I will be merged second")

    util.switch_branch("branchB")
    util.create_file_and_commit("feat: I will be merged first")

    # Note we merge branches opposite order than author_date
    util.switch_branch(main_branch)
    util.merge_branch("branchB")
    util.merge_branch("branchA")

    changelog = Changelog(
        config_customize,
        {"unreleased_version": None, "incremental": True, "dry_run": False},
    )
    mocker.patch.object(changelog.cz, "changelog_hook", changelog_hook_mock)
    changelog()
    full_changelog = "\
## Unreleased\n\n\
\
### Feat\n\n\
- I will be merged second\n\
- I will be merged first\n\
- initial commit\n"

    changelog_hook_mock.assert_called_with(full_changelog, full_changelog)


@pytest.mark.usefixtures("tmp_commitizen_project")
def test_changelog_multiple_incremental_do_not_add_new_lines(
    changelog_path: Path,
    file_regression: FileRegressionFixture,
    util: UtilFixture,
):
    """Test for bug https://github.com/commitizen-tools/commitizen/issues/192"""
    for commit_message in [
        "feat: add new output",
        "fix: output glitch",
        "fix: no more explosions",
        "feat: add more stuff",
    ]:
        util.create_file_and_commit(commit_message)
        util.run_cli("changelog", "--incremental")

    out = changelog_path.read_text(encoding="utf-8")

    file_regression.check(out, extension=".md")


@pytest.mark.usefixtures("tmp_commitizen_project")
def test_changelog_incremental_newline_separates_new_content_from_old(
    changelog_path: Path,
    util: UtilFixture,
    file_regression: FileRegressionFixture,
):
    """Test for https://github.com/commitizen-tools/commitizen/issues/509"""
    with changelog_path.open("w", encoding="utf-8") as f:
        f.write("Pre-existing content that should be kept\n")

    util.create_file_and_commit("feat: add more cat videos")
    util.run_cli("changelog", "--incremental")
    out = changelog_path.read_text(encoding="utf-8")

    file_regression.check(out, extension=".md")


def test_changelog_without_revision(tmp_commitizen_project, util: UtilFixture):
    (tmp_commitizen_project / "CHANGELOG.md").write_text(
        """
        # Unreleased

        ## v1.0.0
        """
    )

    # No revision
    with pytest.raises(NoRevisionError):
        util.run_cli("changelog", "--incremental")

    util.create_file_and_commit("feat: new file")
    util.create_tag("2.0.0")

    # With different tag name and changelog content
    with pytest.raises(NoRevisionError):
        util.run_cli("changelog", "--incremental")


@pytest.mark.usefixtures("tmp_commitizen_project")
def test_changelog_incremental_with_revision(util: UtilFixture):
    """combining incremental with a revision doesn't make sense"""
    with pytest.raises(NotAllowed):
        util.run_cli("changelog", "--incremental", "0.2.0")


@pytest.mark.usefixtures("chdir")
def test_changelog_in_non_git_project(util: UtilFixture):
    with pytest.raises(NotAGitProjectError):
        util.run_cli("changelog", "--incremental")


@pytest.mark.parametrize(
    "commit_message",
    [
        "feat(users): email pattern corrected\n\n"
        "BREAKING CHANGE: migrate by renaming user to users\n\n"
        "footer content",
        "feat(users): email pattern corrected\n\n"
        "body content\n\n"
        "BREAKING CHANGE: migrate by renaming user to users",
    ],
)
@pytest.mark.usefixtures("tmp_commitizen_project")
def test_breaking_change_content_v1_beta(
    capsys: pytest.CaptureFixture,
    file_regression: FileRegressionFixture,
    util: UtilFixture,
    commit_message: str,
):
    util.create_file_and_commit(commit_message)
    with pytest.raises(DryRunExit):
        util.run_cli("changelog", "--dry-run")
    out, _ = capsys.readouterr()
    file_regression.check(out, extension=".md")


@pytest.mark.usefixtures("tmp_commitizen_project")
def test_breaking_change_content_v1_multiline(
    capsys: pytest.CaptureFixture,
    file_regression: FileRegressionFixture,
    util: UtilFixture,
):
    commit_message = (
        "feat(users): email pattern corrected\n\n"
        "body content\n\n"
        "BREAKING CHANGE: migrate by renaming user to users.\n"
        "and then connect the thingy with the other thingy\n\n"
        "footer content"
    )
    util.create_file_and_commit(commit_message)
    with pytest.raises(DryRunExit):
        util.run_cli("changelog", "--dry-run")
    out, _ = capsys.readouterr()
    file_regression.check(out, extension=".md")


@pytest.mark.usefixtures("tmp_commitizen_project")
def test_breaking_change_content_v1_with_exclamation_mark(
    capsys: pytest.CaptureFixture,
    file_regression: FileRegressionFixture,
    util: UtilFixture,
):
    commit_message = "chore!: drop support for py36"
    util.create_file_and_commit(commit_message)
    with pytest.raises(DryRunExit):
        util.run_cli("changelog", "--dry-run")
    out, _ = capsys.readouterr()

    file_regression.check(out, extension=".md")


@pytest.mark.usefixtures("tmp_commitizen_project")
def test_breaking_change_content_v1_with_exclamation_mark_feat(
    capsys: pytest.CaptureFixture,
    file_regression: FileRegressionFixture,
    util: UtilFixture,
):
    commit_message = "feat(pipeline)!: some text with breaking change"
    util.create_file_and_commit(commit_message)
    with pytest.raises(DryRunExit):
        util.run_cli("changelog", "--dry-run")
    out, _ = capsys.readouterr()

    file_regression.check(out, extension=".md")


@pytest.mark.usefixtures("tmp_commitizen_project")
def test_changelog_config_flag_increment(
    changelog_path: Path,
    config_path: Path,
    file_regression: FileRegressionFixture,
    util: UtilFixture,
):
    with config_path.open("a", encoding="utf-8") as f:
        f.write("changelog_incremental = true\n")
    with changelog_path.open("a", encoding="utf-8") as f:
        f.write("\nnote: this should be persisted using increment\n")

    util.create_file_and_commit("feat: add new output")

    util.run_cli("changelog")

    out = changelog_path.read_text(encoding="utf-8")

    assert "this should be persisted using increment" in out
    file_regression.check(out, extension=".md")


@pytest.mark.parametrize("test_input", ["rc", "alpha", "beta"])
@pytest.mark.usefixtures("tmp_commitizen_project")
@pytest.mark.freeze_time("2025-12-29")
def test_changelog_config_flag_merge_prerelease(
    changelog_path: Path,
    config_path: Path,
    file_regression: FileRegressionFixture,
    test_input: str,
    util: UtilFixture,
):
    with config_path.open("a") as f:
        f.write("changelog_merge_prerelease = true\n")

    util.create_file_and_commit("irrelevant commit")
    util.create_tag("1.0.0")

    util.create_file_and_commit("feat: add new output")
    util.create_file_and_commit("fix: output glitch")

    util.run_cli("bump", "--prerelease", test_input, "--yes")

    util.create_file_and_commit("fix: mama gotta work")
    util.create_file_and_commit("feat: add more stuff")
    util.create_file_and_commit("Merge into master")

    util.run_cli("changelog")

    out = changelog_path.read_text()

    file_regression.check(out, extension=".md")


@pytest.mark.usefixtures("tmp_commitizen_project")
def test_changelog_config_start_rev_option(
    capsys: pytest.CaptureFixture,
    config_path: Path,
    file_regression: FileRegressionFixture,
    util: UtilFixture,
):
    # create commit and tag
    util.create_file_and_commit("feat: new file")
    util.run_cli("bump", "--yes")
    capsys.readouterr()

    util.create_file_and_commit("feat: after 0.2.0")
    util.create_file_and_commit("feat: after 0.2")

    with config_path.open("a", encoding="utf-8") as f:
        f.write('changelog_start_rev = "0.2.0"\n')

    with pytest.raises(DryRunExit):
        util.run_cli("changelog", "--dry-run")

    out, _ = capsys.readouterr()
    file_regression.check(out, extension=".md")


@pytest.mark.usefixtures("tmp_commitizen_project")
def test_changelog_incremental_keep_a_changelog_sample_with_annotated_tag(
    changelog_path: Path,
    file_regression: FileRegressionFixture,
    util: UtilFixture,
):
    """Fix #378"""
    with changelog_path.open("w", encoding="utf-8") as f:
        f.write(KEEP_A_CHANGELOG)
    util.create_file_and_commit("irrelevant commit")
    util.create_tag("1.0.0", annotated=True)

    util.create_file_and_commit("feat: add new output")
    util.create_file_and_commit("fix: output glitch")
    util.create_file_and_commit("fix: mama gotta work")
    util.create_file_and_commit("feat: add more stuff")
    util.create_file_and_commit("Merge into master")

    util.run_cli("changelog", "--incremental")

    out = changelog_path.read_text(encoding="utf-8")

    file_regression.check(out, extension=".md")


@pytest.mark.parametrize("test_input", ["rc", "alpha", "beta"])
@pytest.mark.usefixtures("tmp_commitizen_project")
@pytest.mark.freeze_time("2021-06-11")
def test_changelog_incremental_with_release_candidate_version(
    changelog_path: Path,
    file_regression: FileRegressionFixture,
    test_input: str,
    util: UtilFixture,
):
    """Fix #357"""
    with changelog_path.open("w", encoding="utf-8") as f:
        f.write(KEEP_A_CHANGELOG)
    util.create_file_and_commit("irrelevant commit")
    util.create_tag("1.0.0", annotated=True)

    util.create_file_and_commit("feat: add new output")
    util.create_file_and_commit("fix: output glitch")

    util.run_cli("bump", "--changelog", "--prerelease", test_input, "--yes")

    util.create_file_and_commit("fix: mama gotta work")
    util.create_file_and_commit("feat: add more stuff")
    util.create_file_and_commit("Merge into master")

    util.run_cli("changelog", "--incremental")

    out = changelog_path.read_text(encoding="utf-8")

    file_regression.check(out, extension=".md")


@pytest.mark.parametrize(
    ("from_pre", "to_pre"), itertools.product(["alpha", "beta", "rc"], repeat=2)
)
@pytest.mark.usefixtures("tmp_commitizen_project")
@pytest.mark.freeze_time("2021-06-11")
def test_changelog_incremental_with_prerelease_version_to_prerelease_version(
    changelog_path: Path,
    file_regression: FileRegressionFixture,
    from_pre: str,
    to_pre: str,
    util: UtilFixture,
):
    with changelog_path.open("w") as f:
        f.write(KEEP_A_CHANGELOG)
    util.create_file_and_commit("irrelevant commit")
    util.create_tag("1.0.0", annotated=True)

    util.create_file_and_commit("feat: add new output")
    util.create_file_and_commit("fix: output glitch")

    util.run_cli("bump", "--changelog", "--prerelease", from_pre, "--yes")

    util.run_cli("bump", "--changelog", "--prerelease", to_pre, "--yes")

    out = changelog_path.read_text()

    file_regression.check(out, extension=".md")


@pytest.mark.parametrize("test_input", ["rc", "alpha", "beta"])
@pytest.mark.usefixtures("tmp_commitizen_project")
@pytest.mark.freeze_time("2025-12-29")
def test_changelog_release_candidate_version_with_merge_prerelease(
    changelog_path: Path,
    file_regression: FileRegressionFixture,
    test_input: str,
    util: UtilFixture,
):
    """Fix #357"""
    with changelog_path.open("w") as f:
        f.write(KEEP_A_CHANGELOG)
    util.create_file_and_commit("irrelevant commit")
    util.create_tag("1.0.0")

    util.create_file_and_commit("feat: add new output")
    util.create_file_and_commit("fix: output glitch")

    util.run_cli("bump", "--prerelease", test_input, "--yes")

    util.create_file_and_commit("fix: mama gotta work")
    util.create_file_and_commit("feat: add more stuff")
    util.create_file_and_commit("Merge into master")

    util.run_cli("changelog", "--merge-prerelease")

    out = changelog_path.read_text()

    file_regression.check(out, extension=".md")


@pytest.mark.parametrize("test_input", ["rc", "alpha", "beta"])
@pytest.mark.usefixtures("tmp_commitizen_project")
@pytest.mark.freeze_time("2023-04-16")
def test_changelog_incremental_with_merge_prerelease(
    changelog_path: Path,
    file_regression: FileRegressionFixture,
    test_input: str,
    util: UtilFixture,
):
    """Fix #357"""
    with changelog_path.open("w") as f:
        f.write(KEEP_A_CHANGELOG)
    util.create_file_and_commit("irrelevant commit")
    util.create_tag("1.0.0")

    util.create_file_and_commit("feat: add new output")

    util.run_cli("bump", "--prerelease", test_input, "--yes", "--changelog")

    util.create_file_and_commit("fix: output glitch")

    util.run_cli("bump", "--prerelease", test_input, "--yes")

    util.create_file_and_commit("fix: mama gotta work")
    util.create_file_and_commit("feat: add more stuff")
    util.create_file_and_commit("Merge into master")

    util.run_cli("changelog", "--merge-prerelease", "--incremental")

    out = changelog_path.read_text()

    file_regression.check(out, extension=".md")


@pytest.mark.usefixtures("tmp_commitizen_project")
def test_changelog_with_filename_as_empty_string(config_path: Path, util: UtilFixture):
    with config_path.open("a", encoding="utf-8") as f:
        f.write("changelog_file = true\n")

    util.create_file_and_commit("feat: add new output")

    with pytest.raises(NotAllowed):
        util.run_cli("changelog")


@pytest.mark.usefixtures("tmp_commitizen_project")
@pytest.mark.freeze_time("2022-02-13")
def test_changelog_from_rev_first_version_from_arg(
    config_path: Path,
    changelog_path: Path,
    file_regression: FileRegressionFixture,
    util: UtilFixture,
):
    with config_path.open("a", encoding="utf-8") as f:
        f.write('tag_format = "$version"\n')

    # create commit and tag
    util.create_file_and_commit("feat: new file")
    util.run_cli("bump", "--yes")

    util.create_file_and_commit("feat: after 0.2.0")
    util.create_file_and_commit("feat: another feature")

    util.run_cli("bump", "--yes")

    util.run_cli("changelog", "0.2.0")
    out = changelog_path.read_text(encoding="utf-8")

    file_regression.check(out, extension=".md")


@pytest.mark.usefixtures("tmp_commitizen_project")
@pytest.mark.freeze_time("2022-02-13")
def test_changelog_from_rev_latest_version_from_arg(
    config_path: Path,
    changelog_path: Path,
    file_regression: FileRegressionFixture,
    util: UtilFixture,
):
    with config_path.open("a", encoding="utf-8") as f:
        f.write('tag_format = "$version"\n')

    # create commit and tag
    util.create_file_and_commit("feat: new file")
    util.run_cli("bump", "--yes")

    util.create_file_and_commit("feat: after 0.2.0")
    util.create_file_and_commit("feat: another feature")

    util.run_cli("bump", "--yes")

    util.run_cli("changelog", "0.3.0")

    out = changelog_path.read_text()

    file_regression.check(out, extension=".md")


@pytest.mark.usefixtures("tmp_commitizen_project")
@pytest.mark.parametrize(
    ("rev_range", "tag"),
    [
        pytest.param("0.8.0", "0.2.0", id="single-not-found"),
        pytest.param("0.1.0..0.3.0", "0.3.0", id="lower-bound-not-found"),
        pytest.param("0.1.0..0.3.0", "0.1.0", id="upper-bound-not-found"),
        pytest.param("0.3.0..0.4.0", "0.2.0", id="none-found"),
    ],
)
def test_changelog_from_rev_range_not_found(
    config_path: Path, rev_range: str, tag: str, util: UtilFixture
):
    """Provides an invalid revision ID to changelog command"""
    with config_path.open("a", encoding="utf-8") as f:
        f.write('tag_format = "$version"\n')

    # create commit and tag
    util.create_file_and_commit("feat: new file")
    util.create_tag(tag)
    util.create_file_and_commit("feat: new file")
    util.create_tag("1.0.0")

    with pytest.raises(NoCommitsFoundError, match="Could not find a valid revision"):
        util.run_cli("changelog", rev_range)  # it shouldn't exist


@pytest.mark.usefixtures("tmp_commitizen_project")
def test_changelog_multiple_matching_tags(
    config_path: Path, changelog_path: Path, util: UtilFixture
):
    with config_path.open("a", encoding="utf-8") as f:
        f.write('tag_format = "new-$version"\nlegacy_tag_formats = ["legacy-$version"]')

    util.create_file_and_commit("feat: new file")
    util.create_tag("legacy-1.0.0")
    util.create_file_and_commit("feat: new file")
    util.create_tag("legacy-2.0.0")
    util.create_tag("new-2.0.0")

    with pytest.warns() as warnings:
        util.run_cli("changelog", "1.0.0..2.0.0")  # it shouldn't exist

    assert len(warnings) == 1
    warning = warnings[0]
    assert "Multiple tags found for version 2.0.0" in str(warning.message)

    out = changelog_path.read_text()

    # Ensure only one tag is rendered
    assert out.count("2.0.0") == 1


@pytest.mark.usefixtures("tmp_commitizen_project")
def test_changelog_from_rev_range_default_tag_format(
    changelog_path: Path, util: UtilFixture
):
    """Checks that rev_range is calculated with the default (None) tag format"""
    # create commit and tag
    util.create_file_and_commit("feat: new file")
    util.run_cli("bump", "--yes")

    util.create_file_and_commit("feat: after 0.2.0")
    util.create_file_and_commit("feat: another feature")

    util.run_cli("bump", "--yes")

    util.run_cli("changelog", "0.3.0")

    out = changelog_path.read_text()

    assert "new file" not in out


@pytest.mark.usefixtures("tmp_commitizen_project")
@pytest.mark.freeze_time("2022-02-13")
def test_changelog_from_rev_version_range_including_first_tag(
    config_path: Path,
    changelog_path: Path,
    file_regression: FileRegressionFixture,
    util: UtilFixture,
):
    with config_path.open("a", encoding="utf-8") as f:
        f.write('tag_format = "$version"\n')

    # create commit and tag
    util.create_file_and_commit("feat: new file")
    util.run_cli("bump", "--yes")

    util.create_file_and_commit("feat: after 0.2.0")
    util.create_file_and_commit("feat: another feature")

    util.run_cli("bump", "--yes")

    util.run_cli("changelog", "0.2.0..0.3.0")
    out = changelog_path.read_text(encoding="utf-8")

    file_regression.check(out, extension=".md")


@pytest.mark.usefixtures("tmp_commitizen_project")
@pytest.mark.freeze_time("2022-02-13")
def test_changelog_from_rev_version_range_from_arg(
    config_path: Path,
    changelog_path: Path,
    file_regression: FileRegressionFixture,
    util: UtilFixture,
):
    with config_path.open("a", encoding="utf-8") as f:
        f.write('tag_format = "$version"\n')

    # create commit and tag
    util.create_file_and_commit("feat: new file")
    util.run_cli("bump", "--yes")
    util.create_file_and_commit("feat: after 0.2.0")
    util.create_file_and_commit("feat: another feature")

    util.run_cli("bump", "--yes")

    util.create_file_and_commit("feat: getting ready for this")

    util.run_cli("bump", "--yes")

    util.run_cli("changelog", "0.3.0..0.4.0")
    out = changelog_path.read_text()

    file_regression.check(out, extension=".md")


@pytest.mark.usefixtures("tmp_commitizen_project")
@pytest.mark.freeze_time("2022-02-13")
def test_changelog_from_rev_version_range_with_legacy_tags(
    config_path: Path,
    changelog_path: Path,
    file_regression: FileRegressionFixture,
    util: UtilFixture,
):
    Path(config_path).write_text(
        "\n".join(
            [
                "[tool.commitizen]",
                'version_provider = "scm"',
                'tag_format = "v$version"',
                "legacy_tag_formats = [",
                '  "legacy-${version}",',
                '  "old-${version}",',
                "]",
            ]
        ),
    )

    util.create_file_and_commit("feat: new file")
    util.create_tag("old-0.2.0")
    util.create_file_and_commit("feat: new file")
    util.create_tag("legacy-0.3.0")
    util.create_file_and_commit("feat: new file")
    util.create_tag("legacy-0.4.0")

    util.run_cli("changelog", "0.2.0..0.4.0")
    file_regression.check(Path(changelog_path).read_text(), extension=".md")


@pytest.mark.usefixtures("tmp_commitizen_project")
@pytest.mark.freeze_time("2022-02-13")
def test_changelog_from_rev_version_with_big_range_from_arg(
    config_path, changelog_path, file_regression, util: UtilFixture
):
    with config_path.open("a", encoding="utf-8") as f:
        f.write('tag_format = "$version"\n')

    # create commit and tag
    util.create_file_and_commit("feat: new file")

    util.run_cli("bump", "--yes")

    util.create_file_and_commit("feat: after 0.2.0")
    util.create_file_and_commit("feat: another feature")

    util.run_cli("bump", "--yes")  # 0.3.0
    util.create_file_and_commit("feat: getting ready for this")

    util.run_cli("bump", "--yes")  # 0.4.0
    util.create_file_and_commit("fix: small error")

    util.run_cli("bump", "--yes")  # 0.4.1
    util.create_file_and_commit("feat: new shinny feature")

    util.run_cli("bump", "--yes")  # 0.5.0
    util.create_file_and_commit("feat: amazing different shinny feature")

    util.run_cli("bump", "--yes")  # 0.6.0

    util.run_cli("changelog", "0.3.0..0.5.0")
    out = changelog_path.read_text()

    file_regression.check(out, extension=".md")


@pytest.mark.usefixtures("tmp_commitizen_project")
@pytest.mark.freeze_time("2022-02-13")
def test_changelog_from_rev_latest_version_dry_run(
    capsys: pytest.CaptureFixture,
    config_path: Path,
    file_regression: FileRegressionFixture,
    util: UtilFixture,
):
    with config_path.open("a") as f:
        f.write('tag_format = "$version"\n')

    # create commit and tag
    util.create_file_and_commit("feat: new file")
    util.run_cli("bump", "--yes")

    util.create_file_and_commit("feat: after 0.2.0")
    util.create_file_and_commit("feat: another feature")

    util.run_cli("bump", "--yes")
    capsys.readouterr()

    with pytest.raises(DryRunExit):
        util.run_cli("changelog", "0.3.0", "--dry-run")

    out, _ = capsys.readouterr()

    file_regression.check(out, extension=".md")


@pytest.mark.usefixtures("tmp_commitizen_project")
def test_invalid_subject_is_skipped(
    capsys: pytest.CaptureFixture,
    file_regression: FileRegressionFixture,
    util: UtilFixture,
):
    """Fix #510"""
    non_conformant_commit_title = (
        "Merge pull request #487 from manang/master\n\n"
        "feat: skip merge messages that start with Pull request\n"
    )
    util.create_file_and_commit(non_conformant_commit_title)
    util.create_file_and_commit("feat: a new world")
    with pytest.raises(DryRunExit):
        util.run_cli("changelog", "--dry-run")
    out, _ = capsys.readouterr()

    file_regression.check(out, extension=".md")


@pytest.mark.usefixtures("tmp_commitizen_project")
@pytest.mark.freeze_time("2022-02-13")
def test_changelog_with_customized_change_type_order(
    config_path: Path,
    changelog_path: Path,
    file_regression: FileRegressionFixture,
    util: UtilFixture,
):
    with config_path.open("a") as f:
        f.write('tag_format = "$version"\n')
        f.write(
            'change_type_order = ["BREAKING CHANGE", "Perf", "Fix", "Feat", "Refactor"]\n'
        )

    # create commit and tag
    util.create_file_and_commit("feat: new file")
    util.run_cli("bump", "--yes")
    util.create_file_and_commit("feat: after 0.2.0")
    util.create_file_and_commit("feat: another feature")
    util.create_file_and_commit("fix: fix bug")

    util.run_cli("bump", "--yes")

    util.create_file_and_commit("feat: getting ready for this")
    util.create_file_and_commit("perf: perf improvement")

    util.run_cli("bump", "--yes")

    util.run_cli("changelog", "0.3.0..0.4.0")
    out = changelog_path.read_text()

    file_regression.check(out, extension=".md")


@pytest.mark.usefixtures("tmp_commitizen_project")
def test_empty_commit_list(mocker: MockFixture, util: UtilFixture):
    util.create_file_and_commit("feat: a new world")

    # test changelog properly handles when no commits are found for the revision
    mocker.patch("commitizen.git.get_commits", return_value=[])
    with pytest.raises(NoCommitsFoundError):
        util.run_cli("changelog")


@pytest.mark.usefixtures("tmp_commitizen_project")
@pytest.mark.freeze_time("2022-02-13")
def test_changelog_prerelease_rev_with_use_scheme_semver(
    capsys: pytest.CaptureFixture,
    config_path: Path,
    file_regression: FileRegressionFixture,
    util: UtilFixture,
):
    with config_path.open("a") as f:
        f.write('tag_format = "$version"\nversion_scheme = "semver"')

    # create commit and tag
    util.create_file_and_commit("feat: new file")
    util.run_cli("bump", "--yes")

    util.create_file_and_commit("feat: after 0.2.0")
    util.create_file_and_commit("feat: another feature")

    util.run_cli("bump", "--yes", "--prerelease", "alpha")
    capsys.readouterr()

    assert git.tag_exist("0.3.0-a0")

    with pytest.raises(DryRunExit):
        util.run_cli("changelog", "0.3.0-a0", "--dry-run")

    out, _ = capsys.readouterr()
    file_regression.check(out, extension=".md")

    util.run_cli("bump", "--yes", "--prerelease", "alpha")
    capsys.readouterr()

    assert git.tag_exist("0.3.0-a1")

    with pytest.raises(DryRunExit):
        util.run_cli("changelog", "0.3.0-a1", "--dry-run")

    out, _ = capsys.readouterr()
    file_regression.check(out, extension=".second-prerelease.md")


@pytest.mark.usefixtures("tmp_commitizen_project")
def test_changelog_uses_version_tags_for_header(
    mocker: MockFixture, config: BaseConfig, util: UtilFixture
):
    """Tests that changelog headers always use version tags even if there are non-version tags

    This tests a scenario fixed in this commit:
    The first header was using a non-version tag and outputting "## 0-not-a-version" instead of "## 1.0.0
    """
    util.create_file_and_commit("feat: commit in 1.0.0")
    util.create_tag("0-not-a-version")
    util.create_tag("1.0.0")
    util.create_tag("also-not-a-version")

    write_patch = mocker.patch("commitizen.commands.changelog.out.write")

    changelog = Changelog(
        config, {"dry_run": True, "incremental": True, "unreleased_version": None}
    )

    with pytest.raises(DryRunExit):
        changelog()

    changelog_output = write_patch.call_args[0][0]

    assert changelog_output.startswith("## 1.0.0")
    assert "0-no-a-version" not in changelog_output
    assert "also-not-a-version" not in changelog_output


@pytest.mark.usefixtures("tmp_commitizen_project")
@pytest.mark.freeze_time("2022-02-13")
def test_changelog_from_current_version_tag_with_nonversion_tag(
    mocker: MockFixture, config: BaseConfig, util: UtilFixture
):
    """Tests that changelog generation for a single version works even if
    there is a non-version tag in the list of tags

    This tests a scenario which is fixed in this commit:
    You have a commit in between two versions (1.0.0..2.0.0) which is tagged with a non-version tag (not-a-version).
    In this case commitizen should disregard the non-version tag when determining the rev-range & generating the changelog.
    """
    util.create_file_and_commit("feat: initial commit")
    util.create_tag("1.0.0")

    util.create_file_and_commit("feat: commit 1")
    util.create_tag("1-not-a-version")

    util.create_file_and_commit("feat: commit 2")

    util.create_file_and_commit("bump: version 1.0.0 → 2.0.0")
    util.create_tag("2.0.0")

    write_patch = mocker.patch("commitizen.commands.changelog.out.write")

    with pytest.raises(DryRunExit):
        Changelog(
            config,
            {
                "dry_run": True,
                "incremental": False,
                "unreleased_version": None,
                "rev_range": "2.0.0",
            },
        )()

    write_patch.assert_called_with(
        "\
## 2.0.0 (2022-02-13)\n\n\
### Feat\n\n\
- commit 2\n\
- commit 1\n"
    )


@pytest.mark.parametrize(
    ("arg", "cfg", "expected"),
    [
        pytest.param("", "", "default", id="default"),
        pytest.param("", "changelog.cfg", "from config", id="from-config"),
        pytest.param(
            "--template=changelog.cmd", "changelog.cfg", "from cmd", id="from-command"
        ),
    ],
)
def test_changelog_template_option_precedence(
    tmp_commitizen_project: Path,
    arg: str,
    cfg: str,
    expected: str,
    util: UtilFixture,
    changelog_file: Path,
    changelog_tpl: Path,
):
    project_root = tmp_commitizen_project
    cfg_template = project_root / "changelog.cfg"
    cmd_template = project_root / "changelog.cmd"

    cfg_template.write_text("from config")
    cmd_template.write_text("from cmd")
    changelog_tpl.write_text("default")

    util.create_file_and_commit("feat: new file")

    if cfg:
        pyproject = project_root / "pyproject.toml"
        pyproject.write_text(
            dedent(
                f"""\
                [tool.commitizen]
                version = "0.1.0"
                template = "{cfg}"
                """
            )
        )

    testargs = ["changelog"]
    if arg:
        testargs.append(arg)
    util.run_cli(*testargs)

    assert changelog_file.read_text() == expected


def test_changelog_template_extras_precedence(
    changelog_tpl: Path,
    mock_plugin: BaseCommitizen,
    pyproject: Path,
    changelog_file: Path,
    util: UtilFixture,
):
    changelog_tpl.write_text("{{first}} - {{second}} - {{third}}")

    mock_plugin.template_extras = dict(
        first="from-plugin", second="from-plugin", third="from-plugin"
    )

    pyproject.write_text(
        dedent(
            """\
            [tool.commitizen]
            version = "0.1.0"
            [tool.commitizen.extras]
            first = "from-config"
            second = "from-config"
            """
        )
    )

    util.create_file_and_commit("feat: new file")
    util.run_cli("changelog", "--extra", "first=from-command")

    assert changelog_file.read_text() == "from-command - from-config - from-plugin"


@pytest.mark.usefixtures("tmp_commitizen_project")
@pytest.mark.freeze_time("2021-06-11")
def test_changelog_only_tag_matching_tag_format_included_prefix(
    changelog_path: Path,
    config_path: Path,
    util: UtilFixture,
):
    with config_path.open("a", encoding="utf-8") as f:
        f.write('\ntag_format = "custom${version}"\n')
    util.create_file_and_commit("feat: new file")
    util.create_tag("v0.2.0")
    util.create_file_and_commit("feat: another new file")
    util.create_tag("0.2.0")
    util.create_tag("random0.2.0")
    util.run_cli("bump", "--changelog", "--yes")
    util.create_file_and_commit("feat: another new file")
    util.run_cli("bump", "--changelog", "--yes")
    out = changelog_path.read_text()
    assert out.startswith("## custom0.3.0 (2021-06-11)")
    assert "## v0.2.0 (2021-06-11)" not in out
    assert "## 0.2.0  (2021-06-11)" not in out


@pytest.mark.usefixtures("tmp_commitizen_project")
def test_changelog_only_tag_matching_tag_format_included_prefix_sep(
    changelog_path: Path,
    config_path: Path,
    util: UtilFixture,
):
    with config_path.open("a", encoding="utf-8") as f:
        f.write('\ntag_format = "custom-${version}"\n')
    util.create_file_and_commit("feat: new file")
    util.create_tag("v0.2.0")
    util.create_file_and_commit("feat: another new file")
    util.create_tag("0.2.0")
    util.create_tag("random0.2.0")
    util.run_cli("bump", "--changelog", "--yes")
    util.create_file_and_commit("feat: new version another new file")
    util.create_file_and_commit("feat: new version some new file")
    util.run_cli("bump", "--changelog")
    out = changelog_path.read_text()
    assert out.startswith("## custom-0.3.0")
    assert "## v0.2.0" not in out
    assert "## 0.2.0" not in out


@pytest.mark.usefixtures("tmp_commitizen_project")
@pytest.mark.freeze_time("2021-06-11")
def test_changelog_only_tag_matching_tag_format_included_suffix(
    changelog_path: Path,
    config_path: Path,
    util: UtilFixture,
):
    with config_path.open("a", encoding="utf-8") as f:
        f.write('\ntag_format = "${version}custom"\n')
    util.create_file_and_commit("feat: new file")
    util.create_tag("v0.2.0")
    util.create_file_and_commit("feat: another new file")
    util.create_tag("0.2.0")
    util.create_tag("random0.2.0")
    # bump to 0.2.0custom
    util.run_cli("bump", "--changelog", "--yes")

    util.create_file_and_commit("feat: another new file")
    # bump to 0.3.0custom
    util.run_cli("bump", "--changelog", "--yes")
    out = changelog_path.read_text()
    assert out.startswith("## 0.3.0custom (2021-06-11)")
    assert "## v0.2.0 (2021-06-11)" not in out
    assert "## 0.2.0  (2021-06-11)" not in out


@pytest.mark.usefixtures("tmp_commitizen_project")
@pytest.mark.freeze_time("2021-06-11")
def test_changelog_only_tag_matching_tag_format_included_suffix_sep(
    changelog_path: Path,
    config_path: Path,
    util: UtilFixture,
):
    with config_path.open("a", encoding="utf-8") as f:
        f.write('\ntag_format = "${version}-custom"\n')
    util.create_file_and_commit("feat: new file")
    util.create_tag("v0.2.0")
    util.create_file_and_commit("feat: another new file")
    util.create_tag("0.2.0")
    util.create_tag("random0.2.0")
    util.run_cli("bump", "--changelog", "--yes")
    util.create_file_and_commit("feat: another new file")
    util.run_cli("bump", "--changelog", "--yes")
    out = changelog_path.read_text()
    assert out.startswith("## 0.3.0-custom (2021-06-11)")
    assert "## v0.2.0 (2021-06-11)" not in out
    assert "## 0.2.0  (2021-06-11)" not in out


@pytest.mark.usefixtures("tmp_commitizen_project")
def test_changelog_legacy_tags(
    changelog_path: Path,
    config_path: Path,
    util: UtilFixture,
):
    with config_path.open("a", encoding="utf-8") as f:
        f.writelines(
            [
                'tag_format = "v${version}"\n',
                "legacy_tag_formats = [\n",
                '  "older-${version}",\n',
                '  "oldest-${version}",\n',
                "]\n",
            ]
        )
    util.create_file_and_commit("feat: new file")
    util.create_tag("oldest-0.1.0")
    util.create_file_and_commit("feat: new file")
    util.create_tag("older-0.2.0")
    util.create_file_and_commit("feat: another new file")
    util.create_tag("v0.3.0")
    util.create_file_and_commit("feat: another new file")
    util.create_tag("not-0.3.1")
    util.run_cli("bump", "--changelog", "--yes")
    out = changelog_path.read_text()
    assert "## v0.3.0" in out
    assert "## older-0.2.0" in out
    assert "## oldest-0.1.0" in out
    assert "## v0.3.0" in out
    assert "## not-0.3.1" not in out


@pytest.mark.usefixtures("tmp_commitizen_project")
@pytest.mark.freeze_time("2024-11-18")
def test_changelog_incremental_change_tag_format(
    changelog_path: Path,
    config_path: Path,
    file_regression: FileRegressionFixture,
    util: UtilFixture,
):
    util.freezer.move_to("2024-11-18")
    config = Path(config_path)
    base_config = config.read_text()
    config.write_text(
        "\n".join(
            (
                base_config,
                'tag_format = "older-${version}"',
            )
        )
    )
    util.create_file_and_commit("feat: new file")
    util.create_tag("older-0.1.0")
    util.create_file_and_commit("feat: new file")
    util.create_tag("older-0.2.0")
    util.run_cli("changelog")

    config.write_text(
        "\n".join(
            (
                base_config,
                'tag_format = "v${version}"',
                'legacy_tag_formats = ["older-${version}"]',
            )
        )
    )
    util.create_file_and_commit("feat: another new file")
    util.create_tag("v0.3.0")
    util.run_cli("changelog", "--incremental")
    out = changelog_path.read_text()
    assert "## v0.3.0" in out
    assert "## older-0.2.0" in out
    assert "## older-0.1.0" in out
    file_regression.check(out, extension=".md")


@pytest.mark.usefixtures("tmp_commitizen_project")
def test_changelog_ignored_tags(
    changelog_path: Path,
    config_path: Path,
    capsys: pytest.CaptureFixture,
    util: UtilFixture,
):
    with config_path.open("a", encoding="utf-8") as f:
        f.writelines(
            [
                'tag_format = "v${version}"\n',
                "ignored_tag_formats = [\n",
                '  "ignored",\n',
                '  "ignore-${version}",\n',
                "]\n",
            ]
        )
    util.create_file_and_commit("feat: new file")
    util.create_tag("ignore-0.1.0")
    util.create_file_and_commit("feat: new file")
    util.create_tag("ignored")
    util.create_file_and_commit("feat: another new file")
    util.create_tag("v0.3.0")
    util.create_file_and_commit("feat: another new file")
    util.create_tag("not-ignored")
    util.run_cli("bump", "--changelog", "--yes")
    out = changelog_path.read_text()
    assert "## ignore-0.1.0" not in out
    assert "## ignored" not in out
    assert "## not-ignored" not in out
    assert "## v0.3.0" in out

    _, err = capsys.readouterr()
    assert "Invalid version tag: 'ignore-0.1.0'" not in err
    assert "Invalid version tag: 'ignored'" not in err
    assert "Invalid version tag: 'not-ignored'" in err
    assert "Invalid version tag: 'v0.3.0'" not in err


def test_changelog_template_extra_quotes(
    changelog_tpl: Path,
    changelog_file: Path,
    util: UtilFixture,
):
    changelog_tpl.write_text("{{first}} - {{second}} - {{third}}")
    util.create_file_and_commit("feat: new file")
    util.run_cli(
        "changelog",
        "-e",
        "first=no-quote",
        "-e",
        "second='single quotes'",
        "-e",
        'third="double quotes"',
    )

    assert changelog_file.read_text() == "no-quote - single quotes - double quotes"


@pytest.mark.parametrize(
    ("extra", "expected"),
    [
        pytest.param("key=value=", "value=", id="2-equals"),
        pytest.param("key==value", "=value", id="2-consecutive-equals"),
        pytest.param("key==value==", "=value==", id="multiple-equals"),
    ],
)
def test_changelog_template_extra_weird_but_valid(
    changelog_tpl: Path,
    changelog_file: Path,
    extra: str,
    expected: str,
    util: UtilFixture,
):
    changelog_tpl.write_text("{{key}}")
    util.create_file_and_commit("feat: new file")
    util.run_cli("changelog", "-e", extra)

    assert changelog_file.read_text() == expected


@pytest.mark.parametrize("extra", ["no-equal", "", "=no-key"])
def test_changelog_template_extra_bad_format(
    changelog_tpl: Path,
    extra: str,
    util: UtilFixture,
):
    changelog_tpl.write_text("")
    util.create_file_and_commit("feat: new file")
    with pytest.raises(InvalidCommandArgumentError):
        util.run_cli("changelog", "-e", extra)


def test_export_changelog_template_from_default(
    any_changelog_format: ChangelogFormat,
    util: UtilFixture,
    changelog_jinja_file: Path,
    repo_root: Path,
):
    src = repo_root / "commitizen" / "templates" / any_changelog_format.template

    util.run_cli("changelog", "--export-template", str(changelog_jinja_file))

    assert changelog_jinja_file.exists()
    assert changelog_jinja_file.read_text() == src.read_text()


def test_export_changelog_template_from_plugin(
    changelog_jinja_file: Path,
    mock_plugin: BaseCommitizen,
    changelog_format: ChangelogFormat,
    tmp_path: Path,
    util: UtilFixture,
):
    src = tmp_path / changelog_format.template
    tpl = "I am a custom template"
    src.write_text(tpl)
    mock_plugin.template_loader = FileSystemLoader(tmp_path)

    util.run_cli("changelog", "--export-template", str(changelog_jinja_file))

    assert changelog_jinja_file.exists()
    assert changelog_jinja_file.read_text() == tpl


def test_export_changelog_template_fails_when_template_has_no_filename(
    mocker: MockFixture,
    changelog_jinja_file: Path,
    util: UtilFixture,
):
    # Mock a template object with no filename
    class FakeTemplate:
        filename = None

    # Patch get_changelog_template to return a template without a filename
    mocker.patch(
        "commitizen.changelog.get_changelog_template", return_value=FakeTemplate()
    )

    with pytest.raises(NotAllowed) as exc_info:
        util.run_cli("changelog", "--export-template", str(changelog_jinja_file))

    assert not changelog_jinja_file.exists()
    assert "Template filename is not set" in str(exc_info.value)


def test_changelog_template_incremental_variable(
    tmp_commitizen_project: Path,
    any_changelog_format: ChangelogFormat,
    util: UtilFixture,
    file_regression: FileRegressionFixture,
):
    """
    Test that the changelog template is not rendered when the incremental flag is not set.
    Reference: https://github.com/commitizen-tools/commitizen/discussions/1620
    """
    project_root = tmp_commitizen_project
    changelog_tpl = project_root / any_changelog_format.template
    changelog_tpl.write_text(
        dedent("""
        {% if not incremental %}
        # CHANGELOG
        {% endif %}

        {% for entry in tree %}

        ## {{ entry.version }}{% if entry.date %} ({{ entry.date }}){% endif %}

        {% for change_key, changes in entry.changes.items() %}

        {% if change_key %}
        ### {{ change_key }}
        {% endif %}

        {% for change in changes %}
        {% if change.scope %}
        - **{{ change.scope }}**: {{ change.message }}
        {% elif change.message %}
        - {{ change.message }}
        {% endif %}
        {% endfor %}
        {% endfor %}
        {% endfor %}
        """)
    )
    target = "CHANGELOG.md"

    util.create_file_and_commit("feat(foo): new file")
    util.run_cli("changelog", "--file-name", target)
    out = Path(target).read_text(encoding="utf-8")
    file_regression.check(out, extension=".md")

    util.create_file_and_commit("refactor(bar): another new file")
    util.run_cli("changelog", "--file-name", target, "--incremental")
    out = Path(target).read_text(encoding="utf-8")
    file_regression.check(out, extension=".incremental.md")



from __future__ import annotations

import sys
import uuid
from dataclasses import dataclass
from datetime import datetime
from pathlib import Path
from typing import TYPE_CHECKING, NamedTuple

import pytest

from commitizen import cli, cmd, exceptions, git
from commitizen.cmd import Command

if TYPE_CHECKING:
    from unittest.mock import Mock

    from freezegun.api import FrozenDateTimeFactory
    from pytest_mock import MockerFixture

    from commitizen.version_schemes import Increment, Prerelease


class VersionSchemeTestArgs(NamedTuple):
    current_version: str
    increment: Increment | None
    prerelease: Prerelease | None
    prerelease_offset: int
    devrelease: int | None


@dataclass
class UtilFixture:
    """
    An helper fixture providing most common operations for tests.

    git and cli operations grant control over time.
    """

    mocker: MockerFixture
    monkeypatch: pytest.MonkeyPatch
    freezer: FrozenDateTimeFactory

    def __post_init__(self):
        self.patch_env()

    def create_file_and_commit(
        self,
        message: str,
        filename: str | None = None,
        committer_date: str | None = None,
    ) -> None:
        """Create a file and commit it."""
        if not filename:
            filename = str(uuid.uuid4())

        Path(filename).touch()
        c = cmd.run(["git", "add", "."])
        if c.return_code != 0:
            raise exceptions.CommitError(c.err)
        c = git.commit(message, committer_date=committer_date)
        if c.return_code != 0:
            raise exceptions.CommitError(c.err)
        self.tick()

    def create_branch(self, name: str) -> None:
        """Create a new branch."""
        c = cmd.run(["git", "branch", name])
        if c.return_code != 0:
            raise exceptions.GitCommandError(c.err)

    def switch_branch(self, branch: str) -> None:
        """Switch to given branch."""
        c = cmd.run(["git", "switch", branch])
        if c.return_code != 0:
            raise exceptions.GitCommandError(c.err)

    def merge_branch(self, branch: str) -> None:
        """Merge given branch into current branch."""
        c = cmd.run(["git", "merge", branch])
        if c.return_code != 0:
            raise exceptions.GitCommandError(c.err)
        self.tick()

    def get_current_branch(self) -> str:
        """Get current git branch name."""
        c = cmd.run(["git", "rev-parse", "--abbrev-ref", "HEAD"])
        if c.return_code != 0:
            raise exceptions.GitCommandError(c.err)
        return c.out.strip()

    def create_tag(
        self, tag: str, message: str | None = None, annotated: bool | None = None
    ) -> None:
        """Create a git tag."""
        c = git.tag(
            tag, annotated=(annotated is True or message is not None), msg=message
        )
        if c.return_code != 0:
            raise exceptions.CommitError(c.err)
        self.tick()

    def run_cli(self, *args: str) -> None:
        """Execute commitizen CLI with given args."""
        self.mocker.patch.object(sys, "argv", ["cz", *args])
        cli.main()
        self.tick()

    def patch_env(self) -> None:
        """Patch environment variables to sync with FreezeGun time."""
        self.monkeypatch.setenv("DATE", datetime.now().isoformat())
        self.monkeypatch.setenv("GIT_AUTHOR_DATE", datetime.now().isoformat())
        self.monkeypatch.setenv("GIT_COMMITTER_DATE", datetime.now().isoformat())

    def tick(self) -> None:
        """Advance time by 1 second."""
        self.freezer.tick()
        self.patch_env()

    def mock_cmd(self, out: str = "", err: str = "", return_code: int = 0) -> Mock:
        """Mock cmd.run command."""
        return_value = Command(out, err, b"", b"", return_code)
        return self.mocker.patch("commitizen.cmd.run", return_value=return_value)


@pytest.fixture
def util(
    monkeypatch: pytest.MonkeyPatch,
    mocker: MockerFixture,
    freezer: FrozenDateTimeFactory,
) -> UtilFixture:
    """An helper fixture"""
    return UtilFixture(mocker=mocker, monkeypatch=monkeypatch, freezer=freezer)



import platform
import sys

import pytest
from pytest_mock import MockerFixture

from commitizen import commands
from commitizen.__version__ import __version__
from commitizen.config.base_config import BaseConfig


def test_version_for_showing_project_version_error(config, capsys):
    # No version specified in config
    commands.Version(
        config,
        {"project": True},
    )()
    captured = capsys.readouterr()
    assert "No project information in this project." in captured.err


def test_version_for_showing_project_version(config, capsys):
    config.settings["version"] = "0.0.1"
    commands.Version(
        config,
        {"project": True},
    )()
    captured = capsys.readouterr()
    assert "0.0.1" in captured.out


@pytest.mark.parametrize("project", [True, False])
def test_version_for_showing_commitizen_version(config, capsys, project: bool):
    commands.Version(
        config,
        {"project": project, "commitizen": True},
    )()
    captured = capsys.readouterr()
    assert f"{__version__}" in captured.out


def test_version_for_showing_both_versions_no_project(config, capsys):
    commands.Version(
        config,
        {"verbose": True},
    )()
    captured = capsys.readouterr()
    assert f"Installed Commitizen Version: {__version__}" in captured.out
    assert "No project information in this project." in captured.err


def test_version_for_showing_both_versions(config, capsys):
    config.settings["version"] = "0.0.1"
    commands.Version(
        config,
        {"verbose": True},
    )()
    captured = capsys.readouterr()
    expected_out = (
        f"Installed Commitizen Version: {__version__}\nProject Version: 0.0.1"
    )
    assert expected_out in captured.out


def test_version_for_showing_commitizen_system_info(config, capsys):
    commands.Version(
        config,
        {"report": True},
    )()
    captured = capsys.readouterr()
    assert f"Commitizen Version: {__version__}" in captured.out
    assert f"Python Version: {sys.version}" in captured.out
    assert f"Operating System: {platform.system()}" in captured.out


@pytest.mark.parametrize("project", [True, False])
@pytest.mark.usefixtures("tmp_git_project")
def test_version_use_version_provider(
    mocker: MockerFixture,
    config: BaseConfig,
    capsys: pytest.CaptureFixture,
    project: bool,
):
    version = "0.0.0"
    mock = mocker.MagicMock(name="provider")
    mock.get_version.return_value = version
    get_provider = mocker.patch(
        "commitizen.commands.version.get_provider", return_value=mock
    )

    commands.Version(
        config,
        {
            "project": project,
            "verbose": not project,
        },
    )()
    captured = capsys.readouterr()

    assert version in captured.out
    get_provider.assert_called_once()
    mock.get_version.assert_called_once()
    mock.set_version.assert_not_called()


@pytest.mark.parametrize(
    ("version", "expected_version"),
    [
        ("1.0.0", "1\n"),
        ("2.1.3", "2\n"),
        ("0.0.1", "0\n"),
        ("0.1.0", "0\n"),
    ],
)
def test_version_just_major(config, capsys, version: str, expected_version: str):
    config.settings["version"] = version
    commands.Version(
        config,
        {
            "project": True,
            "major": True,
        },
    )()
    captured = capsys.readouterr()
    assert expected_version == captured.out


@pytest.mark.parametrize(
    ("version", "expected_version"),
    [
        ("1.0.0", "0\n"),
        ("2.1.3", "1\n"),
        ("0.0.1", "0\n"),
        ("0.1.0", "1\n"),
    ],
)
def test_version_just_minor(config, capsys, version: str, expected_version: str):
    config.settings["version"] = version
    commands.Version(
        config,
        {
            "project": True,
            "minor": True,
        },
    )()
    captured = capsys.readouterr()
    assert expected_version == captured.out


@pytest.mark.parametrize(
    ("args", "expected_error"),
    [
        (
            {"major": True},
            "can only be used with MANUAL_VERSION, --project or --verbose.",
        ),
        (
            {"minor": True},
            "can only be used with MANUAL_VERSION, --project or --verbose.",
        ),
        (
            {"patch": True},
            "can only be used with MANUAL_VERSION, --project or --verbose.",
        ),
        ({"tag": True}, "Tag can only be used with --project or --verbose."),
    ],
)
def test_version_invalid_combinations(config, capsys, args: dict, expected_error: str):
    """Test that certain flag combinations produce errors."""
    commands.Version(config, args)()  # type: ignore[arg-type]
    captured = capsys.readouterr()
    assert not captured.out
    assert expected_error in captured.err


@pytest.mark.parametrize(
    ("version", "tag_format", "expected_output"),
    [
        ("1.2.3", "v$version", "v1.2.3\n"),
        ("1.2.3", "$version", "1.2.3\n"),
        ("2.0.0", "release-$version", "release-2.0.0\n"),
        ("0.1.0", "ver$version", "ver0.1.0\n"),
    ],
)
def test_version_with_tag_format(
    config, capsys, version: str, tag_format: str, expected_output: str
):
    """Test --tag option applies tag_format to version"""
    config.settings["version"] = version
    config.settings["tag_format"] = tag_format
    commands.Version(
        config,
        {
            "project": True,
            "tag": True,
        },
    )()
    captured = capsys.readouterr()
    assert captured.out == expected_output


@pytest.mark.parametrize(
    ("next_increment", "current_version", "expected_version"),
    [
        ("MAJOR", "1.1.0", "2.0.0"),
        ("MAJOR", "1.0.0", "2.0.0"),
        ("MAJOR", "0.0.1", "1.0.0"),
        ("MINOR", "1.1.0", "1.2.0"),
        ("MINOR", "1.0.0", "1.1.0"),
        ("MINOR", "0.0.1", "0.1.0"),
        ("PATCH", "1.1.0", "1.1.1"),
        ("PATCH", "1.0.0", "1.0.1"),
        ("PATCH", "0.0.1", "0.0.2"),
        ("NONE", "1.0.0", "1.0.0"),
    ],
)
def test_next_version(
    config, capsys, next_increment: str, current_version: str, expected_version: str
):
    config.settings["version"] = current_version
    for project in (True, False):
        commands.Version(
            config,
            {
                "next": next_increment,
                "project": project,
            },
        )()
        captured = capsys.readouterr()
        assert expected_version in captured.out

    # Use the same settings to test the manual version
    commands.Version(
        config,
        {
            "manual_version": current_version,
            "next": next_increment,
        },
    )()
    captured = capsys.readouterr()
    assert expected_version in captured.out


def test_next_version_invalid_version(config, capsys):
    commands.Version(
        config,
        {
            "manual_version": "INVALID",
        },
    )()
    captured = capsys.readouterr()
    assert "Invalid version: 'INVALID'" in captured.err


@pytest.mark.parametrize(
    ("version", "expected_version"),
    [
        ("1.0.0", "0\n"),
        ("2.1.3", "3\n"),
        ("0.0.1", "1\n"),
        ("0.1.0", "0\n"),
    ],
)
def test_version_just_patch(config, capsys, version: str, expected_version: str):
    config.settings["version"] = version
    commands.Version(
        config,
        {
            "project": True,
            "patch": True,
        },
    )()
    captured = capsys.readouterr()
    assert expected_version == captured.out


def test_version_unknown_scheme(config, capsys):
    config.settings["version"] = "1.0.0"
    config.settings["version_scheme"] = "not_a_registered_scheme_name_xyz"
    commands.Version(config, {"project": True})()
    captured = capsys.readouterr()
    assert "Unknown version scheme." in captured.err


def test_version_use_git_commits_not_implemented(config, capsys):
    config.settings["version"] = "1.0.0"
    commands.Version(
        config,
        {"project": True, "next": "USE_GIT_COMMITS"},
    )()
    captured = capsys.readouterr()
    assert "USE_GIT_COMMITS is not implemented" in captured.err


def test_version_no_arguments_shows_commitizen_version(config, capsys):
    commands.Version(config, {})()
    captured = capsys.readouterr()
    assert captured.out.strip() == __version__



import subprocess
from pathlib import Path


def gen_cli_interactive_gifs() -> None:
    """Generate GIF screenshots for interactive commands using VHS."""
    vhs_dir = Path(__file__).parent.parent / "docs" / "images"
    output_dir = Path(__file__).parent.parent / "docs" / "images" / "cli_interactive"
    output_dir.mkdir(parents=True, exist_ok=True)

    vhs_files = list(vhs_dir.glob("*.tape"))

    if not vhs_files:
        print("No VHS tape files found in docs/images/, skipping")
        return

    for vhs_file in vhs_files:
        print(f"Processing: {vhs_file.name}")
        try:
            subprocess.run(
                ["vhs", vhs_file.name],
                check=True,
                cwd=vhs_dir,
            )
            gif_name = vhs_file.stem + ".gif"
            print(f"✓ Generated {gif_name}")
        except FileNotFoundError:
            print(
                "✗ VHS is not installed. Please install it from: "
                "https://github.com/charmbracelet/vhs"
            )
            raise
        except subprocess.CalledProcessError as e:
            print(f"✗ Error processing {vhs_file.name}: {e}")
            raise


if __name__ == "__main__":
    gen_cli_interactive_gifs()



from __future__ import annotations

import os
import re
import subprocess
import sys
import tempfile
from pathlib import Path
from typing import TYPE_CHECKING, cast

import pytest

from commitizen import cmd, defaults
from commitizen.changelog_formats import (
    ChangelogFormat,
    get_changelog_format,
)
from commitizen.config import BaseConfig
from commitizen.cz import registry
from commitizen.cz.base import BaseCommitizen

if TYPE_CHECKING:
    from collections.abc import Iterator, Mapping

    from pytest_mock import MockerFixture

    from commitizen.question import CzQuestion
    from tests.utils import UtilFixture


SIGNER = "GitHub Action"
SIGNER_MAIL = "action@github.com"
PYTHON_VERSION = f"{sys.version_info.major}.{sys.version_info.minor}"

pytest_plugins = [
    "tests.utils",
]


@pytest.fixture
def repo_root() -> Path:
    return Path(__file__).parent.parent


@pytest.fixture
def in_repo_root(repo_root: Path) -> Iterator[Path]:
    cwd = os.getcwd()
    os.chdir(repo_root)
    yield repo_root
    os.chdir(cwd)


@pytest.fixture
def data_dir(repo_root: Path) -> Path:
    return repo_root / "tests" / "data"


@pytest.fixture(scope="session")
def set_default_gitconfig() -> dict[str, str]:
    return {
        "user.name": "SIGNER",
        "user.email": SIGNER_MAIL,
        "safe.directory": "*",
        "init.defaultBranch": "master",
    }


@pytest.fixture
def chdir(tmp_path: Path) -> Iterator[Path]:
    cwd = os.getcwd()
    os.chdir(tmp_path)
    yield tmp_path
    os.chdir(cwd)


@pytest.fixture
def tmp_git_project(tmp_path: Path, monkeypatch: pytest.MonkeyPatch):
    monkeypatch.chdir(tmp_path)
    cmd.run(["git", "init"])

    return tmp_path


@pytest.fixture
def tmp_commitizen_project(tmp_git_project: Path):
    (tmp_git_project / "pyproject.toml").write_text(
        '[tool.commitizen]\nversion="0.1.0"\n'
    )

    return tmp_git_project


@pytest.fixture
def pyproject(tmp_commitizen_project: Path) -> Path:
    return tmp_commitizen_project / "pyproject.toml"


@pytest.fixture
def tmp_commitizen_project_initial(
    tmp_git_project: Path, util: UtilFixture, monkeypatch: pytest.MonkeyPatch
):
    def _initial(
        config_extra: str | None = None,
        version="0.1.0",
        initial_commit="feat: new user interface",
    ):
        monkeypatch.chdir(tmp_git_project)
        tmp_commitizen_cfg_file = tmp_git_project / "pyproject.toml"
        tmp_commitizen_cfg_file.write_text(f'[tool.commitizen]\nversion="{version}"\n')
        tmp_version_file = tmp_git_project / "__version__.py"
        tmp_version_file.write_text(version)
        tmp_version_file_string = str(tmp_version_file).replace("\\", "/")
        with tmp_commitizen_cfg_file.open("a", encoding="utf-8") as f:
            f.write(f'\nversion_files = ["{tmp_version_file_string}"]\n')
        if config_extra:
            with tmp_commitizen_cfg_file.open("a", encoding="utf-8") as f:
                f.write(config_extra)
        util.create_file_and_commit(initial_commit)

        return tmp_git_project

    return _initial


def _get_gpg_keyid(signer_mail):
    _new_key = cmd.run(["gpg", "--list-secret-keys", signer_mail])
    _m = re.search(
        r"[a-zA-Z0-9 \[\]-_]*\n[ ]*([0-9A-Za-z]*)\n[\na-zA-Z0-9 \[\]-_<>@]*",
        _new_key.out,
    )
    return _m.group(1) if _m else None


@pytest.fixture
def tmp_commitizen_project_with_gpg(tmp_commitizen_project):
    # create a temporary GPGHOME to store a temporary keyring.
    # Home path must be less than 104 characters
    gpg_home = tempfile.TemporaryDirectory(suffix="_cz")
    old_gnupghome = os.environ.get("GNUPGHOME")
    if os.name != "nt":
        os.environ["GNUPGHOME"] = gpg_home.name  # tempdir = temp keyring

    try:
        # create a key (a keyring will be generated within GPUPGHOME)
        subprocess.run(
            [
                "gpg",
                "--batch",
                "--yes",
                "--debug-quick-random",
                "--passphrase",
                "",
                "--quick-gen-key",
                f"{SIGNER} {SIGNER_MAIL}",
            ],
            check=True,
        )
        key_id = _get_gpg_keyid(SIGNER_MAIL)
        assert key_id

        # configure git to use gpg signing
        cmd.run(["git", "config", "commit.gpgsign", "true"])
        cmd.run(["git", "config", "user.signingkey", key_id])

        yield tmp_commitizen_project
    finally:
        if old_gnupghome is not None:
            os.environ["GNUPGHOME"] = old_gnupghome
        elif "GNUPGHOME" in os.environ and os.name != "nt":
            os.environ.pop("GNUPGHOME")
        gpg_home.cleanup()


@pytest.fixture
def config():
    _config = BaseConfig()
    _config.settings.update({"name": defaults.DEFAULT_SETTINGS["name"]})
    return _config


class SemverCommitizen(BaseCommitizen):
    """A minimal cz rules used to test changelog and bump.

    Samples:
    ```
    minor(users): add email to user
    major: removed user profile
    patch(deps): updated dependency for security
    ```
    """

    bump_pattern = r"^(patch|minor|major)"
    bump_map = {
        "major": "MAJOR",
        "minor": "MINOR",
        "patch": "PATCH",
    }
    bump_map_major_version_zero = {
        "major": "MINOR",
        "minor": "MINOR",
        "patch": "PATCH",
    }
    changelog_pattern = r"^(patch|minor|major)"
    commit_parser = r"^(?P<change_type>patch|minor|major)(?:\((?P<scope>[^()\r\n]*)\)|\()?:?\s(?P<message>.+)"
    change_type_map = {
        "major": "Breaking Changes",
        "minor": "Features",
        "patch": "Bugs",
    }

    def questions(self) -> list[CzQuestion]:
        return [
            {
                "type": "list",
                "name": "prefix",
                "message": "Select the type of change you are committing",
                "choices": [
                    {
                        "value": "patch",
                        "name": "patch: a bug fix",
                        "key": "p",
                    },
                    {
                        "value": "minor",
                        "name": "minor: a new feature, non-breaking",
                        "key": "m",
                    },
                    {
                        "value": "major",
                        "name": "major: a breaking change",
                        "key": "b",
                    },
                ],
            },
            {
                "type": "input",
                "name": "subject",
                "message": (
                    "Write a short and imperative summary of the code changes: (lower case and no period)\n"
                ),
            },
        ]

    def message(self, answers: Mapping) -> str:
        prefix = answers["prefix"]
        subject = answers.get("subject", "default message").trim()
        return f"{prefix}: {subject}"

    def example(self) -> str:
        return ""

    def schema(self) -> str:
        return ""

    def schema_pattern(self) -> str:
        return ""

    def info(self) -> str:
        return ""


@pytest.fixture
def use_cz_semver(mocker):
    new_cz = {**registry, "cz_semver": SemverCommitizen}
    mocker.patch.dict("commitizen.cz.registry", new_cz)


class MockPlugin(BaseCommitizen):
    def questions(self) -> list[CzQuestion]:
        return []

    def message(self, answers: Mapping) -> str:
        return ""

    def example(self) -> str:
        return ""

    def schema(self) -> str:
        return ""

    def schema_pattern(self) -> str:
        return ""

    def info(self) -> str:
        return ""


@pytest.fixture
def mock_plugin(mocker: MockerFixture, config: BaseConfig) -> BaseCommitizen:
    mock = MockPlugin(config)
    mocker.patch("commitizen.factory.committer_factory", return_value=mock)
    return mock


SUPPORTED_FORMATS = ("markdown", "textile", "asciidoc", "restructuredtext")


@pytest.fixture(params=SUPPORTED_FORMATS)
def changelog_format(
    config: BaseConfig, request: pytest.FixtureRequest
) -> ChangelogFormat:
    """For tests relying on formats specifics"""
    format: str = request.param
    config.settings["changelog_format"] = format
    if "tmp_commitizen_project" in request.fixturenames:
        tmp_commitizen_project = request.getfixturevalue("tmp_commitizen_project")
        pyproject = tmp_commitizen_project / "pyproject.toml"
        with pyproject.open("a", encoding="utf-8") as f:
            f.write(f'\nchangelog_format = "{format}"\n')
    return get_changelog_format(config)


@pytest.fixture
def any_changelog_format(config: BaseConfig) -> ChangelogFormat:
    """For test not relying on formats specifics, use the default"""
    config.settings["changelog_format"] = defaults.CHANGELOG_FORMAT
    return get_changelog_format(config)


@pytest.fixture(params=[pytest.param(PYTHON_VERSION, id=f"py_{PYTHON_VERSION}")])
def python_version(request: pytest.FixtureRequest) -> str:
    """The current python version in '{major}.{minor}' format"""
    return cast("str", request.param)


@pytest.fixture
def consistent_terminal_output(monkeypatch: pytest.MonkeyPatch):
    """Force consistent terminal output."""
    monkeypatch.setenv("COLUMNS", "80")
    monkeypatch.setenv("TERM", "dumb")
    monkeypatch.setenv("LC_ALL", "C")
    monkeypatch.setenv("LANG", "C")
    monkeypatch.setenv("NO_COLOR", "1")
    monkeypatch.setenv("PAGER", "cat")



from __future__ import annotations

from enum import IntEnum


class VersionIncrement(IntEnum):
    """Semantic versioning bump increments.

    IntEnum keeps a total order compatible with NONE < PATCH < MINOR < MAJOR
    for comparisons across the codebase.

    - NONE: no bump (docs-only / style commits, etc.)
    - PATCH: backwards-compatible bug fixes
    - MINOR: backwards-compatible features
    - MAJOR: incompatible API changes
    """

    NONE = 0
    PATCH = 1
    MINOR = 2
    MAJOR = 3

    def __str__(self) -> str:
        return self.name

    @classmethod
    def from_value(cls, value: object) -> VersionIncrement:
        if not isinstance(value, str):
            return VersionIncrement.NONE
        try:
            return cls[value]
        except KeyError:
            return VersionIncrement.NONE



import re
import sys
from pathlib import Path
from unittest.mock import ANY

import pytest
from pytest_mock import MockFixture, MockType

from commitizen import cli, cmd, commands
from commitizen.cz.exceptions import CzException
from commitizen.cz.utils import get_backup_file_path
from commitizen.exceptions import (
    CommitError,
    CommitMessageLengthExceededError,
    CustomError,
    DryRunExit,
    NoAnswersError,
    NoCommitBackupError,
    NotAGitProjectError,
    NotAllowed,
    NothingToCommitError,
)


@pytest.fixture
def commit_mock(mocker: MockFixture):
    return mocker.patch(
        "commitizen.git.commit", return_value=cmd.Command("success", "", b"", b"", 0)
    )


@pytest.fixture
def prompt_mock_feat(mocker: MockFixture):
    return mocker.patch(
        "questionary.prompt",
        return_value={
            "prefix": "feat",
            "subject": "user created",
            "scope": "",
            "is_breaking_change": False,
            "body": "closes #21",
            "footer": "",
        },
    )


@pytest.fixture
def staging_is_clean(mocker: MockFixture, tmp_git_project):
    mocker.patch("commitizen.git.is_staging_clean", return_value=False)
    return tmp_git_project


@pytest.fixture
def backup_file(tmp_git_project, monkeypatch):
    """Write backup message so Commit finds it when run from tmp_git_project."""
    monkeypatch.chdir(tmp_git_project)
    path = get_backup_file_path()
    path.write_text("backup commit", encoding="utf-8")


@pytest.mark.usefixtures("staging_is_clean", "commit_mock", "prompt_mock_feat")
def test_commit(config, success_mock: MockType):
    commands.Commit(config, {})()
    success_mock.assert_called_once()


@pytest.mark.usefixtures("staging_is_clean")
def test_commit_backup_on_failure(
    config, mocker: MockFixture, prompt_mock_feat: MockType
):
    mocker.patch(
        "commitizen.git.commit", return_value=cmd.Command("", "error", b"", b"", 9)
    )
    error_mock = mocker.patch("commitizen.out.error")

    commit_cmd = commands.Commit(config, {})
    temp_file = commit_cmd.backup_file_path
    with pytest.raises(CommitError):
        commit_cmd()

    prompt_mock_feat.assert_called_once()
    error_mock.assert_called_once()
    assert Path(temp_file).exists()


@pytest.mark.usefixtures("staging_is_clean", "commit_mock")
def test_commit_retry_fails_no_backup(config):
    with pytest.raises(
        NoCommitBackupError, match=re.escape(NoCommitBackupError.message)
    ):
        commands.Commit(config, {"retry": True})()


@pytest.mark.usefixtures("staging_is_clean", "backup_file")
def test_commit_retry_works(
    config, success_mock: MockType, mocker: MockFixture, commit_mock: MockType
):
    prompt_mock = mocker.patch("questionary.prompt")

    commit_cmd = commands.Commit(config, {"retry": True})
    temp_file = commit_cmd.backup_file_path
    commit_cmd()

    commit_mock.assert_called_with("backup commit", args=[])
    prompt_mock.assert_not_called()
    success_mock.assert_called_once()
    assert not Path(temp_file).exists()


@pytest.mark.usefixtures("staging_is_clean")
def test_commit_retry_after_failure_no_backup(
    config, success_mock: MockType, commit_mock: MockType, prompt_mock_feat: MockType
):
    config.settings["retry_after_failure"] = True
    commands.Commit(config, {})()

    commit_mock.assert_called_with("feat: user created\n\ncloses #21", args=[])
    prompt_mock_feat.assert_called_once()
    success_mock.assert_called_once()


@pytest.mark.usefixtures("staging_is_clean", "backup_file")
def test_commit_retry_after_failure_works(
    config, success_mock: MockType, mocker: MockFixture, commit_mock: MockType
):
    prompt_mock = mocker.patch("questionary.prompt")

    config.settings["retry_after_failure"] = True
    commit_cmd = commands.Commit(config, {})
    temp_file = commit_cmd.backup_file_path
    commit_cmd()

    commit_mock.assert_called_with("backup commit", args=[])
    prompt_mock.assert_not_called()
    success_mock.assert_called_once()
    assert not Path(temp_file).exists()


@pytest.mark.usefixtures("staging_is_clean", "backup_file")
def test_commit_retry_after_failure_with_no_retry_works(
    config, success_mock: MockType, commit_mock: MockType, prompt_mock_feat: MockType
):
    config.settings["retry_after_failure"] = True
    commit_cmd = commands.Commit(config, {"no_retry": True})
    temp_file = commit_cmd.backup_file_path
    commit_cmd()

    commit_mock.assert_called_with("feat: user created\n\ncloses #21", args=[])
    prompt_mock_feat.assert_called_once()
    success_mock.assert_called_once()
    assert not Path(temp_file).exists()


@pytest.mark.usefixtures("staging_is_clean", "prompt_mock_feat")
def test_commit_command_with_dry_run_option(config):
    with pytest.raises(DryRunExit):
        commands.Commit(config, {"dry_run": True})()


@pytest.mark.usefixtures("staging_is_clean", "commit_mock", "prompt_mock_feat")
def test_commit_command_with_write_message_to_file_option(
    config, tmp_path, success_mock: MockType
):
    tmp_file = tmp_path / "message"
    commands.Commit(config, {"write_message_to_file": tmp_file})()
    success_mock.assert_called_once()
    assert tmp_file.exists()
    assert "feat: user created" in tmp_file.read_text()


@pytest.mark.usefixtures("staging_is_clean", "prompt_mock_feat")
def test_commit_command_with_invalid_write_message_to_file_option(config, tmp_path):
    with pytest.raises(NotAllowed):
        commands.Commit(config, {"write_message_to_file": tmp_path})()


@pytest.mark.usefixtures("staging_is_clean", "prompt_mock_feat")
def test_commit_command_with_signoff_option(
    config, success_mock: MockType, commit_mock: MockType
):
    commands.Commit(config, {"signoff": True})()

    commit_mock.assert_called_once_with(ANY, args=["-s"])
    success_mock.assert_called_once()


@pytest.mark.usefixtures("staging_is_clean", "prompt_mock_feat")
def test_commit_command_with_always_signoff_enabled(
    config, success_mock: MockType, commit_mock: MockType
):
    config.settings["always_signoff"] = True
    commands.Commit(config, {})()

    commit_mock.assert_called_once_with(ANY, args=["-s"])
    success_mock.assert_called_once()


@pytest.mark.usefixtures("staging_is_clean", "prompt_mock_feat")
def test_commit_command_with_gpgsign_and_always_signoff_enabled(
    config, success_mock: MockType, commit_mock: MockType
):
    config.settings["always_signoff"] = True
    commands.Commit(config, {"extra_cli_args": ["-S"]})()

    commit_mock.assert_called_once_with(ANY, args=["-S", "-s"])
    success_mock.assert_called_once()


@pytest.mark.usefixtures("tmp_git_project")
def test_commit_when_nothing_to_commit(config, mocker: MockFixture):
    mocker.patch("commitizen.git.is_staging_clean", return_value=True)

    with pytest.raises(NothingToCommitError, match="No files added to staging!"):
        commands.Commit(config, {})()


@pytest.mark.usefixtures("staging_is_clean", "prompt_mock_feat")
def test_commit_with_allow_empty(config, success_mock: MockType, commit_mock: MockType):
    commands.Commit(config, {"extra_cli_args": ["--allow-empty"]})()
    commit_mock.assert_called_with(
        "feat: user created\n\ncloses #21", args=["--allow-empty"]
    )
    success_mock.assert_called_once()


@pytest.mark.usefixtures("staging_is_clean", "prompt_mock_feat")
def test_commit_with_signoff_and_allow_empty(
    config, success_mock: MockType, commit_mock: MockType
):
    config.settings["always_signoff"] = True
    commands.Commit(config, {"extra_cli_args": ["--allow-empty"]})()

    commit_mock.assert_called_with(
        "feat: user created\n\ncloses #21", args=["--allow-empty", "-s"]
    )
    success_mock.assert_called_once()


@pytest.mark.usefixtures("staging_is_clean")
def test_commit_when_customized_expected_raised(config, mocker: MockFixture):
    _err = ValueError()
    _err.__context__ = CzException("This is the root custom err")
    mocker.patch("questionary.prompt", side_effect=_err)
    with pytest.raises(CustomError, match="This is the root custom err"):
        commands.Commit(config, {})()


@pytest.mark.usefixtures("staging_is_clean")
def test_commit_when_non_customized_expected_raised(config, mocker: MockFixture):
    mocker.patch("questionary.prompt", side_effect=ValueError("error message"))
    with pytest.raises(ValueError, match="error message"):
        commands.Commit(config, {})()


@pytest.mark.usefixtures("staging_is_clean")
def test_commit_when_no_user_answer(config, mocker: MockFixture):
    mocker.patch("questionary.prompt", return_value=None)
    with pytest.raises(NoAnswersError):
        commands.Commit(config, {})()


def test_commit_in_non_git_project(tmp_path, monkeypatch, config):
    monkeypatch.chdir(tmp_path)
    with pytest.raises(NotAGitProjectError):
        commands.Commit(config, {})


@pytest.mark.usefixtures("staging_is_clean", "commit_mock", "prompt_mock_feat")
def test_commit_command_with_all_option(
    config, success_mock: MockType, mocker: MockFixture
):
    add_mock = mocker.patch("commitizen.git.add")
    commands.Commit(config, {"all": True})()
    add_mock.assert_called()
    success_mock.assert_called_once()


@pytest.mark.usefixtures("staging_is_clean", "prompt_mock_feat")
def test_commit_command_with_extra_args(
    config, success_mock: MockType, commit_mock: MockType
):
    commands.Commit(config, {"extra_cli_args": ["--", "-extra-args1", "-extra-arg2"]})()
    commit_mock.assert_called_once_with(ANY, args=["--", "-extra-args1", "-extra-arg2"])
    success_mock.assert_called_once()


@pytest.mark.usefixtures(
    "tmp_commitizen_project", "staging_is_clean", "prompt_mock_feat"
)
def test_commit_cli_extra_args_parsing(mocker: MockFixture, commit_mock: MockType):
    """Test that extra args after -- are passed as a list through the CLI."""
    mocker.patch.object(sys, "argv", ["cz", "commit", "--", "--no-verify"])
    cli.main()
    commit_mock.assert_called_once_with(ANY, args=["--no-verify"])


@pytest.mark.usefixtures("staging_is_clean")
@pytest.mark.parametrize("editor", ["vim", None])
def test_manual_edit(editor, config, mocker: MockFixture, tmp_path):
    mocker.patch("commitizen.git.get_core_editor", return_value=editor)
    subprocess_mock = mocker.patch("subprocess.call")

    mocker.patch("shutil.which", return_value=editor)

    test_message = "Initial commit message"
    temp_file = tmp_path / "temp_commit_message"
    temp_file.write_text(test_message)

    mock_temp_file = mocker.patch("tempfile.NamedTemporaryFile")
    mock_temp_file.return_value.__enter__.return_value.name = str(temp_file)

    commit_cmd = commands.Commit(config, {"edit": True})

    if editor is None:
        with pytest.raises(RuntimeError):
            commit_cmd.manual_edit(test_message)
    else:
        edited_message = commit_cmd.manual_edit(test_message)
        subprocess_mock.assert_called_once_with(["vim", str(temp_file)])
        assert edited_message == test_message.strip()


@pytest.mark.usefixtures("staging_is_clean", "prompt_mock_feat")
@pytest.mark.parametrize(
    "out", ["no changes added to commit", "nothing added to commit"]
)
def test_commit_when_nothing_added_to_commit(config, mocker: MockFixture, out):
    commit_mock = mocker.patch(
        "commitizen.git.commit",
        return_value=cmd.Command(
            out=out,
            err="",
            stdout=out.encode(),
            stderr=b"",
            return_code=0,
        ),
    )
    error_mock = mocker.patch("commitizen.out.error")

    commands.Commit(config, {})()

    commit_mock.assert_called_once()
    error_mock.assert_called_once_with(out)


@pytest.mark.usefixtures("staging_is_clean", "commit_mock")
def test_commit_command_with_config_message_length_limit(
    config, success_mock: MockType, prompt_mock_feat: MockType
):
    prefix = prompt_mock_feat.return_value["prefix"]
    subject = prompt_mock_feat.return_value["subject"]
    message_length = len(f"{prefix}: {subject}")

    commands.Commit(config, {"message_length_limit": message_length})()
    success_mock.assert_called_once()

    with pytest.raises(CommitMessageLengthExceededError):
        commands.Commit(config, {"message_length_limit": message_length - 1})()

    config.settings["message_length_limit"] = message_length
    success_mock.reset_mock()
    commands.Commit(config, {})()
    success_mock.assert_called_once()

    config.settings["message_length_limit"] = message_length - 1
    with pytest.raises(CommitMessageLengthExceededError):
        commands.Commit(config, {})()

    # Test config message length limit is overridden by CLI argument
    success_mock.reset_mock()
    commands.Commit(config, {"message_length_limit": message_length})()
    success_mock.assert_called_once()

    success_mock.reset_mock()
    commands.Commit(config, {"message_length_limit": 0})()
    success_mock.assert_called_once()



import os
import subprocess
from itertools import chain
from pathlib import Path

from rich.console import Console

from commitizen.cli import data


def gen_cli_help_screenshots() -> None:
    """Generate the screenshot for help message on each cli command and save them as svg files."""
    images_root = Path(__file__).parent.parent / "docs" / "images" / "cli_help"
    images_root.mkdir(parents=True, exist_ok=True)

    cz_commands = (
        command["name"] if isinstance(command["name"], str) else command["name"][0]
        for command in data["subcommands"]["commands"]  # type: ignore[index]
    )
    for cmd in chain(
        ["cz --help"], (f"cz {cz_command} --help" for cz_command in cz_commands)
    ):
        file_name = f"{cmd.replace(' ', '_').replace('-', '_')}.svg"
        _export_cmd_as_svg(cmd, images_root / file_name)


def _export_cmd_as_svg(cmd: str, file_path: Path) -> None:
    console = Console(record=True, width=80, file=open(os.devnull, "w"))

    print("Processing command:", cmd)

    console.print(f"$ {cmd}")
    stdout = subprocess.run(cmd, shell=True, capture_output=True).stdout.decode("utf-8")
    console.print(stdout)
    console.save_svg(file_path.as_posix(), title="")

    print("Saved to:", file_path.as_posix())


if __name__ == "__main__":
    gen_cli_help_screenshots()



import platform
import sys
from typing import TypedDict

from packaging.version import InvalidVersion

from commitizen import out
from commitizen.__version__ import __version__
from commitizen.config import BaseConfig
from commitizen.exceptions import NoVersionSpecifiedError, VersionSchemeUnknown
from commitizen.providers import get_provider
from commitizen.tags import TagRules
from commitizen.version_increment import VersionIncrement
from commitizen.version_schemes import Increment, get_version_scheme


class VersionArgs(TypedDict, total=False):
    manual_version: str | None
    next: str | None

    # Exclusive groups 1
    commitizen: bool
    report: bool
    project: bool
    verbose: bool

    # Exclusive groups 2
    major: bool
    minor: bool
    patch: bool
    tag: bool


class Version:
    """Get the version of the installed commitizen or the current project.
    Precedence:
    1. report
    2. commitizen
    3. verbose, project
    """

    def __init__(self, config: BaseConfig, arguments: VersionArgs) -> None:
        self.config: BaseConfig = config
        self.arguments = arguments

    def __call__(self) -> None:
        if self.arguments.get("report"):
            out.write(f"Commitizen Version: {__version__}")
            out.write(f"Python Version: {sys.version}")
            out.write(f"Operating System: {platform.system()}")
            return

        if self.arguments.get("verbose"):
            out.write(f"Installed Commitizen Version: {__version__}")

        if self.arguments.get("commitizen"):
            out.write(__version__)
            return

        if (
            self.arguments.get("project")
            or self.arguments.get("verbose")
            or self.arguments.get("next")
            or self.arguments.get("manual_version")
        ):
            version_str = self.arguments.get("manual_version")
            if version_str is None:
                try:
                    version_str = get_provider(self.config).get_version()
                except NoVersionSpecifiedError:
                    out.error("No project information in this project.")
                    return
            try:
                scheme_factory = get_version_scheme(self.config.settings)
            except VersionSchemeUnknown:
                out.error("Unknown version scheme.")
                return

            try:
                version = scheme_factory(version_str)
            except InvalidVersion:
                out.error(f"Invalid version: '{version_str}'")
                return

            if next_increment_str := self.arguments.get("next"):
                if next_increment_str == "USE_GIT_COMMITS":
                    # TODO: implement USE_GIT_COMMITS by deriving the increment from
                    # git history. This requires refactoring the bump logic out of
                    # `commitizen/commands/bump.py` so it can be reused here. See #1678.
                    out.error("--next USE_GIT_COMMITS is not implemented yet.")
                    return

                next_increment = VersionIncrement.from_value(next_increment_str)
                increment: Increment | None
                if next_increment == VersionIncrement.NONE:
                    increment = None
                elif next_increment == VersionIncrement.PATCH:
                    increment = "PATCH"
                elif next_increment == VersionIncrement.MINOR:
                    increment = "MINOR"
                else:
                    increment = "MAJOR"
                version = version.bump(increment=increment)

            if self.arguments.get("major"):
                out.write(version.major)
                return
            if self.arguments.get("minor"):
                out.write(version.minor)
                return
            if self.arguments.get("patch"):
                out.write(version.micro)
                return

            display_version: str
            if self.arguments.get("tag"):
                tag_rules = TagRules.from_settings(self.config.settings)
                display_version = tag_rules.normalize_tag(version)
            else:
                display_version = str(version)

            out.write(
                f"Project Version: {display_version}"
                if self.arguments.get("verbose")
                else display_version
            )
            return

        for argument in ("major", "minor", "patch"):
            if self.arguments.get(argument):
                out.error(
                    f"{argument} can only be used with MANUAL_VERSION, --project or --verbose."
                )
                return

        if self.arguments.get("tag"):
            out.error("Tag can only be used with --project or --verbose.")
            return

        # If no arguments are provided, just show the installed commitizen version
        out.write(__version__)
